@@ -18,6 +18,9 @@ from trytond.wizard import (Wizard, StateView, StateTransition, StateReport,
     Button)
 from trytond.tools import grouped_slice
 from trytond.modules.company import CompanyReport
+from trytond.modules.currency.fields import Monetary
+
+from .exceptions import PrintDepreciationTableError
 
 
 def date2datetime(date):
@@ -96,59 +99,53 @@ class Asset(Workflow, ModelSQL, ModelView):
         required=True)
     currency = fields.Function(fields.Many2One('currency.currency',
         'Currency'), 'on_change_with_currency')
-    currency_digits = fields.Function(fields.Integer('Currency Digits'),
-        'on_change_with_currency_digits')
-    quantity = fields.Float('Quantity',
-        digits=(16, Eval('unit_digits', 2)),
+    quantity = fields.Float(
+        "Quantity", digits='unit',
         states={
             'readonly': (Bool(Eval('supplier_invoice_line', 1))
                 | Eval('lines', [0])
                 | (Eval('state') != 'draft')),
             },
-        depends=['state', 'unit_digits'])
+        depends=['state'])
     unit = fields.Many2One('product.uom', 'Unit',
         states={
             'readonly': (Bool(Eval('product'))
                 | (Eval('state') != 'draft')),
             },
         depends=['state'])
-    unit_digits = fields.Function(fields.Integer('Unit Digits'),
-        'on_change_with_unit_digits')
-    value = fields.Numeric('Value',
-        digits=(16, Eval('currency_digits', 2)),
+    value = Monetary(
+        "Value", currency='currency', digits='currency',
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
-        depends=['currency_digits', 'state'],
+        depends=['state'],
         required=True,
         help="The value of the asset when purchased.")
-    depreciated_amount = fields.Numeric("Depreciated Amount",
-        digits=(16, Eval('currency_digits', 2)),
+    depreciated_amount = Monetary(
+        "Depreciated Amount", currency='currency', digits='currency',
         domain=[
             ('depreciated_amount', '<=', Eval('value')),
             ],
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
-        depends=['currency_digits', 'value', 'state'],
+        depends=['value', 'state'],
         required=True,
         help="The amount already depreciated at the start date.")
-    depreciating_value = fields.Function(fields.Numeric(
-            "Depreciating Value",
-            digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits'],
+    depreciating_value = fields.Function(Monetary(
+            "Depreciating Value", currency='currency', digits='currency',
             help="The value of the asset at the start date."),
         'on_change_with_depreciating_value')
-    residual_value = fields.Numeric('Residual Value',
+    residual_value = Monetary(
+        "Residual Value", currency='currency', digits='currency',
+        required=True,
         domain=[
             ('residual_value', '<=', Eval('depreciating_value')),
             ],
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
-        depends=['currency_digits', 'depreciating_value', 'state'],
-        required=True,
-        digits=(16, Eval('currency_digits', 2)))
+        depends=['depreciating_value', 'state'])
     purchase_date = fields.Date('Purchase Date', states={
             'readonly': (Bool(Eval('supplier_invoice_line', 1))
                 | Eval('lines', [0])
@@ -208,6 +205,8 @@ class Asset(Workflow, ModelSQL, ModelView):
             },
         depends=['company'])
     comment = fields.Text('Comment')
+    revisions = fields.One2Many(
+        'account.asset.revision', 'asset', "Revisions", readonly=True)
 
     @classmethod
     def __setup__(cls):
@@ -220,8 +219,14 @@ class Asset(Workflow, ModelSQL, ModelView):
         cls._transitions |= set((
                 ('draft', 'running'),
                 ('running', 'closed'),
+                ('running', 'draft'),
                 ))
         cls._buttons.update({
+                'draft': {
+                    'invisible': (Eval('lines', [])
+                        | (Eval('state') != 'running')),
+                    'depends': ['state'],
+                    },
                 'run': {
                     'invisible': Eval('state') != 'draft',
                     'depends': ['state'],
@@ -231,13 +236,12 @@ class Asset(Workflow, ModelSQL, ModelView):
                     'depends': ['state'],
                     },
                 'create_lines': {
-                    'invisible': (Eval('lines', [])
-                        | (Eval('state') != 'draft')),
+                    'invisible': ~Eval('state').in_(['draft', 'running']),
                     'depends': ['state'],
                     },
                 'clear_lines': {
                     'invisible': (~Eval('lines', [0])
-                        | (Eval('state') != 'draft')),
+                        | ~Eval('state').in_(['draft', 'running'])),
                     'depends': ['state'],
                     },
                 'update': {
@@ -285,6 +289,11 @@ class Asset(Workflow, ModelSQL, ModelView):
     def default_company():
         return Transaction().context.get('company')
 
+    @fields.depends('company')
+    def on_change_company(self):
+        self.frequency = self.default_frequency(
+            company=self.company.id if self.company else None)
+
     @staticmethod
     def default_account_journal():
         Journal = Pool().get('account.journal')
@@ -307,12 +316,6 @@ class Asset(Workflow, ModelSQL, ModelView):
         if self.company:
             return self.company.currency.id
 
-    @fields.depends('company')
-    def on_change_with_currency_digits(self, name=None):
-        if self.company:
-            return self.company.currency.digits
-        return 2
-
     @fields.depends('supplier_invoice_line', 'unit')
     def on_change_supplier_invoice_line(self):
         pool = Pool()
@@ -355,12 +358,6 @@ class Asset(Workflow, ModelSQL, ModelView):
             return None
         return self.product.default_uom.id
 
-    @fields.depends('unit')
-    def on_change_with_unit_digits(self, name=None):
-        if not self.unit:
-            return 2
-        return self.unit.digits
-
     @fields.depends('end_date', 'product', 'start_date')
     def on_change_with_end_date(self):
         if (all(getattr(self, k, None) for k in ('product', 'start_date'))
@@ -400,14 +397,20 @@ class Asset(Workflow, ModelSQL, ModelView):
         # dateutil >= 2.0 has replace __nonzero__ by __bool__ which doesn't
         # work in Python < 3
         if delta == relativedelta.relativedelta():
-            return [self.end_date]
+            if not self.lines:
+                return [self.end_date]
+            else:
+                return []
         if self.frequency == 'monthly':
             rule = rrule.rrule(rrule.MONTHLY, dtstart=self.start_date,
-                bymonthday=int(config.asset_bymonthday))
+                bymonthday=int(config.get_multivalue(
+                        'asset_bymonthday', company=self.company.id)))
         elif self.frequency == 'yearly':
             rule = rrule.rrule(rrule.YEARLY, dtstart=self.start_date,
-                bymonth=int(config.asset_bymonth),
-                bymonthday=int(config.asset_bymonthday))
+                bymonth=int(config.get_multivalue(
+                            'asset_bymonth', company=self.company.id)),
+                bymonthday=int(config.get_multivalue(
+                            'asset_bymonthday', company=self.company.id)))
         dates = [d.date()
             for d in rule.between(date2datetime(start_date),
                 date2datetime(self.end_date))]
@@ -500,7 +503,6 @@ class Asset(Workflow, ModelSQL, ModelView):
     def create_lines(cls, assets):
         pool = Pool()
         Line = pool.get('account.asset.line')
-        cls.clear_lines(assets)
 
         lines = []
         for asset in assets:
@@ -567,6 +569,8 @@ class Asset(Workflow, ModelSQL, ModelView):
         Move = pool.get('account.move')
         Line = pool.get('account.asset.line')
 
+        cls.create_lines(assets)
+
         moves = []
         lines = []
         for asset_ids in grouped_slice(assets):
@@ -583,7 +587,7 @@ class Asset(Workflow, ModelSQL, ModelView):
         Line.save(lines)
         Move.post(moves)
 
-    def get_closing_move(self, account):
+    def get_closing_move(self, account, date=None):
         """
         Returns closing move values.
         """
@@ -593,7 +597,8 @@ class Asset(Workflow, ModelSQL, ModelView):
         Move = pool.get('account.move')
         MoveLine = pool.get('account.move.line')
 
-        date = Date.today()
+        if date is None:
+            date = Date.today()
         period_id = Period.find(self.company.id, date)
         if self.supplier_invoice_line:
             account_asset = self.supplier_invoice_line.account.current()
@@ -636,16 +641,26 @@ class Asset(Workflow, ModelSQL, ModelView):
         Fill the number field with asset sequence.
         '''
         pool = Pool()
-        Sequence = pool.get('ir.sequence')
         Config = pool.get('account.configuration')
 
         config = Config(1)
         for asset in assets:
             if asset.number:
                 continue
-            asset.number = Sequence.get_id(config.asset_sequence.id)
+            asset.number = config.get_multivalue(
+                'asset_sequence', company=asset.company.id).get()
         cls.save(assets)
 
+    @classmethod
+    @ModelView.button
+    @Workflow.transition('draft')
+    def draft(cls, assets):
+        for asset in assets:
+            if asset.lines:
+                raise AccessError(
+                    gettext('account_asset.msg_draft_lines',
+                        asset=asset.rec_name))
+
     @classmethod
     @ModelView.button
     @Workflow.transition('running')
@@ -656,7 +671,7 @@ class Asset(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('closed')
-    def close(cls, assets, account=None):
+    def close(cls, assets, account=None, date=None):
         """
         Close the assets.
         If account is provided, it will be used instead of the expense account.
@@ -666,7 +681,7 @@ class Asset(Workflow, ModelSQL, ModelView):
         cls.clear_lines(assets)
         moves = []
         for asset in assets:
-            moves.append(asset.get_closing_move(account))
+            moves.append(asset.get_closing_move(account, date=date))
         Move.save(moves)
         for move, asset in zip(moves, assets):
             asset.move = move
@@ -691,9 +706,11 @@ class Asset(Workflow, ModelSQL, ModelView):
         else:
             default = default.copy()
         default.setdefault('lines', [])
+        default.setdefault('update_moves', [])
         default.setdefault('number', None)
         default.setdefault('supplier_invoice_line', None)
         default.setdefault('move')
+        default.setdefault('revisions', [])
         return super(Asset, cls).copy(assets, default=default)
 
     @classmethod
@@ -712,27 +729,28 @@ class AssetLine(ModelSQL, ModelView):
     asset = fields.Many2One('account.asset', 'Asset', required=True,
         ondelete='CASCADE', readonly=True)
     date = fields.Date('Date', readonly=True)
-    depreciation = fields.Numeric('Depreciation',
-        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'],
+    depreciation = Monetary(
+        "Depreciation", currency='currency', digits='currency',
         required=True, readonly=True)
-    acquired_value = fields.Numeric('Acquired Value', readonly=True,
-        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
-    depreciable_basis = fields.Numeric('Depreciable Basis', readonly=True,
-        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
-    actual_value = fields.Numeric('Actual Value', readonly=True,
-        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
-    accumulated_depreciation = fields.Numeric(
-        'Accumulated Depreciation', readonly=True,
-        digits=(16, Eval('currency_digits', 2)), depends=['currency_digits'])
+    acquired_value = Monetary(
+        "Acquired Value", currency='currency', digits='currency',
+        readonly=True)
+    depreciable_basis = Monetary(
+        "Depreciable Basis", currency='currency', digits='currency',
+        readonly=True)
+    actual_value = Monetary(
+        "Actual Value", currency='currency', digits='currency', readonly=True)
+    accumulated_depreciation = Monetary(
+        "Accumulated Depreciation", currency='currency', digits='currency',
+        readonly=True)
     move = fields.Many2One('account.move', 'Account Move', readonly=True)
     currency = fields.Function(fields.Many2One('currency.currency',
         'Currency'), 'on_change_with_currency')
-    currency_digits = fields.Function(fields.Integer('Currency Digits'),
-        'on_change_with_currency_digits')
 
     @classmethod
     def __setup__(cls):
         super(AssetLine, cls).__setup__()
+        cls.__access__.add('asset')
         cls._order.insert(0, ('date', 'ASC'))
 
     @fields.depends('asset', '_parent_asset.currency')
@@ -740,11 +758,6 @@ class AssetLine(ModelSQL, ModelView):
         if self.asset:
             return self.asset.currency.id
 
-    @fields.depends('asset', '_parent_asset.currency_digits')
-    def on_change_with_currency_digits(self, name=None):
-        if self.asset:
-            return self.asset.currency_digits
-
 
 class AssetUpdateMove(ModelSQL):
     'Asset - Update - Move'
@@ -777,21 +790,15 @@ class CreateMoves(Wizard):
 
     def transition_create_moves(self):
         Asset = Pool().get('account.asset')
-        assets = Asset.search([
-                ('state', '=', 'running'),
-                ])
+        with Transaction().set_context(_check_access=True):
+            assets = Asset.search([
+                    ('state', '=', 'running'),
+                    ])
+        assets = Asset.browse(assets)
         Asset.create_moves(assets, self.start.date)
         return 'end'
 
 
-class UpdateAssetStart(ModelView):
-    'Update Asset Start'
-    __name__ = 'account.asset.update.start'
-    value = fields.Numeric('Asset Value', required=True)
-    residual_value = fields.Numeric('Residual Value', required=True)
-    end_date = fields.Date('End Date', required=True)
-
-
 class UpdateAssetShowDepreciation(ModelView):
     'Update Asset Show Depreciation'
     __name__ = 'account.asset.update.show_depreciation'
@@ -816,8 +823,8 @@ class UpdateAssetShowDepreciation(ModelView):
 class UpdateAsset(Wizard):
     'Update Asset'
     __name__ = 'account.asset.update'
-    start = StateView('account.asset.update.start',
-        'account_asset.asset_update_start_view_form', [
+    start = StateView('account.asset.revision',
+        'account_asset.asset_revision_view_form', [
             Button('Cancel', 'end', 'tryton-cancel'),
             Button('OK', 'update_asset', 'tryton-ok', True),
             ])
@@ -835,6 +842,7 @@ class UpdateAsset(Wizard):
             'value': self.record.value,
             'residual_value': self.record.residual_value,
             'end_date': self.record.end_date,
+            'asset': self.record.id,
             }
 
     def transition_update_asset(self):
@@ -919,17 +927,70 @@ class UpdateAsset(Wizard):
                 'residual_value': self.start.residual_value,
                 'end_date': self.start.end_date,
                 })
+        self.model.clear_lines([self.record])
         self.model.create_lines([self.record])
+        self.start.asset = self.record
+        self.start.save()
         return 'end'
 
 
+class AssetRevision(ModelSQL, ModelView):
+    "Asset Revision"
+    __name__ = 'account.asset.revision'
+    currency = fields.Function(
+        fields.Many2One('currency.currency', "Currency"),
+        'on_change_with_currency')
+    value = Monetary(
+        "Asset Value", currency='currency', digits='currency',
+        required=True)
+    residual_value = Monetary(
+        "Residual Value", currency='currency', digits='currency',
+        required=True)
+    end_date = fields.Date("End Date", required=True)
+    origin = fields.Reference("Origin", selection='get_origins', select=True)
+    description = fields.Char("Description")
+    asset = fields.Many2One(
+        'account.asset', "Asset", select=True, required=True)
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('asset')
+
+    @fields.depends('asset', '_parent_asset.currency')
+    def on_change_with_currency(self, name=None):
+        if self.asset and self.asset.currency:
+            return self.asset.currency.id
+
+    @fields.depends('origin', 'value', 'asset', '_parent_asset.value')
+    def on_change_origin(self, name=None):
+        pool = Pool()
+        InvoiceLine = pool.get('account.invoice.line')
+        if isinstance(self.origin, InvoiceLine) and self.origin.id >= 0:
+            self.value = self.asset.value + self.origin.amount
+
+    @staticmethod
+    def _get_origin():
+        "Return list of Model names for origin Reference"
+        return ['account.invoice.line']
+
+    @classmethod
+    def get_origins(cls):
+        pool = Pool()
+        IrModel = pool.get('ir.model')
+
+        get_name = IrModel.get_name
+        models = cls._get_origin()
+        return [(None, '')] + [(m, get_name(m)) for m in models]
+
+
 class AssetDepreciationTable(CompanyReport):
     'Asset Depreciation Table'
     __name__ = 'account.asset.depreciation_table'
 
     @classmethod
-    def get_context(cls, records, data):
-        context = super(AssetDepreciationTable, cls).get_context(records, data)
+    def get_context(cls, records, header, data):
+        context = super().get_context(records, header, data)
 
         AssetDepreciation = cls.get_asset_depreciation()
         AssetDepreciation.start_date = data['start_date']
@@ -1004,7 +1065,9 @@ class AssetDepreciationTable(CompanyReport):
 
             @cached_property
             def start_fixed_value(self):
-                if (self.start_date < self.asset.start_date
+                if self.asset.end_date < self.start_date:
+                    return self.asset.value
+                elif (self.start_date < self.asset.start_date
                         or not self.asset_lines):
                     return 0
                 value = self.asset_lines[0].acquired_value
@@ -1031,7 +1094,7 @@ class AssetDepreciationTable(CompanyReport):
             @cached_property
             def end_fixed_value(self):
                 if not self.asset_lines:
-                    return 0
+                    return self.start_fixed_value
                 value = self.asset_lines[-1].acquired_value
                 date = self.asset_lines[-1].date
                 for line in self.update_lines:
@@ -1042,7 +1105,10 @@ class AssetDepreciationTable(CompanyReport):
             @cached_property
             def start_value(self):
                 if not self.asset_lines:
-                    return self.asset.value
+                    if self.asset.start_date > self.end_date:
+                        return self.asset.value
+                    else:
+                        return 0
                 return (self.asset_lines[0].actual_value
                     + self.asset_lines[0].depreciation)
 
@@ -1059,7 +1125,7 @@ class AssetDepreciationTable(CompanyReport):
             @cached_property
             def end_value(self):
                 if not self.asset_lines:
-                    return self.asset.value
+                    return self.start_value
                 return self.asset_lines[-1].actual_value
 
             @cached_property
@@ -1115,11 +1181,21 @@ class PrintDepreciationTable(Wizard):
     def do_print_(self, action):
         pool = Pool()
         Asset = pool.get('account.asset')
-        assets = Asset.search([
-                ('start_date', '<', self.start.end_date),
-                ('end_date', '>', self.start.start_date),
-                ('state', '!=', 'draft'),
-                ])
+        with Transaction().set_context(_check_access=True):
+            assets = Asset.search([
+                    ['OR',
+                        ('purchase_date', '<=', self.start.end_date),
+                        ('start_date', '<=', self.start.end_date),
+                        ],
+                    ['OR',
+                        ('move', '=', None),
+                        ('move.date', '>=', self.start.start_date),
+                        ],
+                    ('state', '!=', 'draft'),
+                    ])
+        if not assets:
+            raise PrintDepreciationTableError(
+                gettext('account_asset.msg_no_assets'))
         return action, {
             'ids': [a.id for a in assets],
             'start_date': self.start.start_date,
